/* wptCardEdit.cpp - OpenPGP editing interface
 *	Copyright (C) 2005, 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "gpgme.h"
#include "wptCard.h"
#include "wptCardEdit.h"
#include "wptGPG.h"
#include "wptErrors.h"

/* Create a new Smartcard context. */
static gpgme_error_t
gpg_card_new (gpg_card_t *ret_card)
{
    gpg_card_t c;

    if (!ret_card)
	return gpg_error (GPG_ERR_INV_ARG);
    *ret_card = NULL;
    c = (gpg_card_t)calloc (1, sizeof *c);
    if (!c)
	return gpg_error (GPG_ERR_ENOMEM);
    *ret_card = c;
    return 0;
}


/* Release an existing Smartcard context. @card. */
void
gpg_card_release (gpg_card_t card)
{
    int i;

    if (!card)
	return;
    if (card->aid)
	free (card->aid);   
    if (card->version)
	free (card->version);
    if (card->lang)
	free (card->lang);
    if (card->login)
	free (card->login);
    for (i=0; i < 3; i++) {
	if (card->fpr[i])
	    free (card->fpr[i]);
	if (card->ca_fpr[i])
	    free (card->ca_fpr[i]);
	if (card->fpr_created_str[i])
	    free (card->fpr_created_str[i]);
    }
    if (card->surname)
	free (card->surname);
    if (card->givenname)
	free (card->givenname);
    if (card->serial)
	free (card->serial);
    if (card->vendor)
	free (card->vendor);
    if (card->url)
	free (card->url);
    if (card->card_type)
	free (card->card_type);
    free (card);

}


/* Parse the timestamp @long_ts and return a string
   representation of it (format: YYYY-DD-MM HH:MM:SS). */
static char*
get_str_timestamp (long long_ts)
{
    struct tm *tm;
    char *p, *fmt;

    tm = localtime (&long_ts);
    fmt = "%04d-%02d-%02d %02d:%02d:%02d";
    p = (char*)calloc (1, strlen (fmt)+32);
    if (!p)
	abort ();
    sprintf (p, fmt,
	     tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
	     tm->tm_hour, tm->tm_min, tm->tm_sec);
    return p;
}


/* Colon handler for parsing the GPG card colon output. */
static gpgme_error_t
statuscard_colon_handler (gpg_card_t card, char * line)
{
    enum rectype_t {
	CARD_None = 0,
	CARD_AID,
	CARD_Version,
	CARD_Vendor,
	CARD_Serial,
	CARD_Name,
	CARD_Lang,
	CARD_Sex,
	CARD_Url,
	CARD_Login,
	CARD_MaxPinLen,
	CARD_SigCount,
	CARD_CAFpr,
	CARD_Fpr,
	CARD_FprTime
    };
    enum rectype_t rectype = (rectype_t)0;
    char *p, *pend;
    int field = 0;

    if (!line)
	return gpg_error (GPG_ERR_EOF);

    for (p = line; p; p = pend) {
	field++;
	pend = strchr (p, ':');
        if (pend)
            *pend++ = 0;

	if (field == 1) {
	    if (!strcmp (p, "AID"))
		rectype = CARD_AID;
	    else if( !strcmp( p, "version" ) )
		rectype = CARD_Version;
	    else if( !strcmp( p, "vendor" ) )
		rectype = CARD_Vendor;
	    else if( !strcmp( p, "serial" ) )
		rectype = CARD_Serial;
	    else if( !strcmp( p, "name" ) )
		rectype = CARD_Name;
	    else if( !strcmp( p, "lang" ) )
		rectype = CARD_Lang;
	    else if( !strcmp( p, "sex" ) )
		rectype = CARD_Sex;
	    else if( !strcmp( p, "url" ) )
		rectype = CARD_Url;
	    else if( !strcmp( p, "login" ) )
		rectype = CARD_Login;
	    else if( !strcmp( p, "maxpinlen" ) )
		rectype = CARD_MaxPinLen;
	    else if( !strcmp( p, "sigcount" ) )
		rectype = CARD_SigCount;
	    else if (!strcmp (p, "cafpr"))
		rectype = CARD_CAFpr;
	    else if (!strcmp (p, "fpr"))
		rectype = CARD_Fpr;	    
	    else if (!strcmp (p, "fprtime"))
		rectype = CARD_FprTime;
	    else
		rectype = CARD_None;
	}
	switch (rectype) {
	case CARD_AID:
	    if (field == 2) {
		card->aid = strdup (p);
		if (!card->aid)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    else if (field == 3) {
		card->card_type = strdup (p);
		if (!card->card_type)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;

	case CARD_Version:
	    if (field == 2) {
		card->version = strdup (p);
		if (!card->version)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;

	case CARD_Vendor:
	    if (field == 3) {
		card->vendor = strdup (p);
		if (!card->vendor)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;

	case CARD_Serial:
	    if (field == 2) {
		card->serial = strdup (p);
		if (!card->serial)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;
	case CARD_Name:
	    if (field == 2) {
		card->givenname = strdup (p);
		if (!card->givenname)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    else if (field == 3) {
		card->surname = strdup (p);
		if (!card->surname)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;
	case CARD_Lang:
	    if( field == 2 ) {
		card->lang = strdup (p);
		if (!card->lang)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;
	case CARD_Sex:
	    if( field == 2 )
		card->sex = *p;
	    break;

	case CARD_Url:
	    if( field == 2 ) {
		if (card->url)
		    free (card->url);
		card->url = (char*)calloc (1, strlen (p) + 1);
		if( !card->url ) {
		    return gpg_error (GPG_ERR_ENOMEM);
		}
		gpg_decode_c_string (p, &card->url, strlen (p) + 1);
		if (!card->url)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;

	case CARD_Login:
	    if (field == 2) {
		card->login = strdup (p);
		if (!card->login)
		    return gpg_error (GPG_ERR_ENOMEM);
	    }
	    break;

	case CARD_MaxPinLen:
	    break;

	case CARD_SigCount:
	    if (field == 2)
		card->sig_count = atol (p);
	    break;

	case CARD_CAFpr:
	    if (!p)
		break;
	    if (field > 4 || field == 1)
		break;
	    card->ca_fpr[field-2] = strdup (p);
	    if (!card->ca_fpr[field-2])
		return gpg_error (GPG_ERR_ENOMEM);
	    break;

	case CARD_Fpr:
	    if (field > 4 || field == 1) /* max 3 fprs */
		break;
	    card->fpr[field-2] = strdup (p);
	    if (!card->fpr[field-2])
		return gpg_error (GPG_ERR_ENOMEM);
	    break;

	case CARD_FprTime:
	    if (field > 4 || field == 1)
		break;
	    card->fpr_created[field-2] = strtoul (p, NULL, 10);
	    card->fpr_created_str[field-2] = get_str_timestamp (strtoul (p, NULL, 10));
	    if (!card->fpr_created_str[field-2])
		return gpg_error (GPG_ERR_ENOMEM);
	    break;
	    
	default:
	    break;
	}
    }
    return 0;
}


/* Dummy handler to get the colon data and then quit. */
static gpgme_error_t
list_handler (void *opaque, gpgme_status_code_t code, const char *key, int fd)
{
    static int wait_card=0;
    GpgCardEdit *ce = (GpgCardEdit *)opaque;
    const char *s="";    
    DWORD n;
        

    if (code == GPGME_STATUS_CARDCTRL) {
	if (!strcmp (key, "5"))
	    ce->setResult (GPG_CARDRES_NOCARD);
	else if (!strcmp (key, "1"))
	    wait_card = 1;
    }
    if (wait_card && !strcmp (key, "cardctrl.insert_card.okay")) {
	n = MessageBox (NULL, _("Please insert the card and click OK or Cancel to abort."),
			_("GPG Card Status"), MB_ICONQUESTION|MB_OKCANCEL);
	if (n == IDCANCEL) {
	    ce->setResult (GPG_CARDRES_CANCEL);
	    WriteFile ((HANDLE)fd, "c\n", 2, &n, NULL);
	    wait_card = 0;
	}
    }
    if (!strcmp (key, "cardedit.prompt")) {
	s = "quit\n";
	if (!WriteFile ((HANDLE)fd, s, strlen (s), &n, NULL))
	    log_debug ("list_handler: WriteFile() failed ec=%d\r\n", 
		       (int)GetLastError ());
	wait_card = 0;
    }
    return 0;
}


/* Construct an empty object. */
GpgCardEdit::GpgCardEdit (void)
{
    pin = NULL;
    pin_new = NULL;
    admin_pin = NULL;

    keygen.comment = NULL;
    keygen.email = NULL;
    keygen.name = NULL;
    keygen.expdate = NULL;
    keygen.key_fpr = NULL;

    type = 0;
    cnt = 0;
    cancel = 0;
    result = 0;

    gpgme_new (&ctx); /* XXX: check return code */
}

/* Release the object. */
GpgCardEdit::~GpgCardEdit (void)
{    
    if (keygen.name)
	free (keygen.name);
    if (keygen.comment)
	free (keygen.comment);
    if (keygen.email)
	free (keygen.email);
    if (keygen.expdate)
	free (keygen.expdate);
    if (keygen.key_fpr)
	free (keygen.key_fpr);
    gpgme_release (ctx);
}


/* Set the user PIN for the object to @pin. */
void
GpgCardEdit::setPIN (const char *_pin)
{
    this->pin = _pin;
}

/* Set the admin PIN for the object to @admin_pin. */
void
GpgCardEdit::setAdminPIN (const char *_admin_pin)
{
    this->admin_pin = _admin_pin;
}


/* Set the new user PIN for the object to @new_pin. */
void 
GpgCardEdit::setNewPIN (const char *_new_pin)
{
    this->pin_new = _new_pin;
}


/* Set the passphrase needed when generating a key to @pass. */
void
GpgCardEdit::setKeygenPassphrase (const char *_pass)
{
    this->keygen.pass = _pass;
}


/* Set the callback to @cb and the hook value to @cb_value. */
void
GpgCardEdit::setCallback (const char* (*_cb)(int code, void *opaque),
			  void *_cb_value)
{
    this->cb_value = _cb_value;
    this->card_cb = _cb;
}


/* Read the information from the inserted card and return
   it in form of a card context @r_card.
   Return value: 0 on success. */
gpgme_error_t
GpgCardEdit::getCardStatus (gpg_card_t *r_card)
{
    gpgme_data_t out = NULL;
    gpgme_error_t err;
    gpg_card_t card=NULL;
    char buf[200];

    if (!r_card)
	return gpg_error (GPG_ERR_INV_ARG);

    err = gpg_card_new (&card);
    if (err)
	return err;

    err = gpgme_data_new (&out);
    if (err) {
	gpg_card_release (card);
	return err;
    }
    
    err = gpgme_op_card_edit  (ctx, NULL, list_handler, this, out);
    gpgme_data_rewind (out);
    if (err) {
	gpgme_data_release (out);
	gpg_card_release (card);
	return err;
    }
    if (getResult () & GPG_CARDRES_NOCARD ||
	getResult () & GPG_CARDRES_CANCEL) {
	gpg_card_release (card);
	gpgme_data_release (out);
	return gpg_error (GPG_ERR_CARD_NOT_PRESENT);
    }

    *r_card = card;
    while (gpg_data_readline (out, buf, sizeof (buf)-2) > 0) {
	err = statuscard_colon_handler (card, buf);
	if (err)
	    break;
    }
    gpgme_data_release (out);

    return err;
}


/* Generate a key on a smart card with the following params:
   @flags: user specific params.
   @name: name of the key holder.
   @email: email of the key holder.
   @comment: optional comment.
   @valid: how long is the key valid in days.
   Return value: 0 on success. */
gpgme_error_t
GpgCardEdit::genKey (int flags, const char *name,
		     const char *email, const char *comment, long valid,
		     char **r_key_fpr)
{
    gpgme_error_t err;

    if (!name || !email)
	return gpg_error (GPG_ERR_INV_ARG);
    if (!this->keygen.pass)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITCARD_GENKEY;
    this->keygen.flags = flags;

    if (this->keygen.name)
	free (this->keygen.name);
    this->keygen.name = strdup (name);
    if (this->keygen.email)
	free (this->keygen.email);
    this->keygen.email = strdup (email);
    if (this->keygen.comment)
	free (this->keygen.comment);
    this->keygen.comment = comment? strdup (comment) : strdup ("");
    if (this->keygen.expdate)
	free (this->keygen.expdate);
    this->keygen.expdate = (char*)calloc (1, 16);
    if (this->keygen.expdate)
	sprintf (this->keygen.expdate, "%lu", valid);

    if (!this->keygen.name || !this->keygen.email)
	return gpg_error (GPG_ERR_ENOMEM);

    err = gpg_card_edit (ctx, this);
    if (!err)
	err = getResult () & GPG_CARDRES_CANCEL? gpg_error (GPG_ERR_CANCELED) : 0;
    if (!err && r_key_fpr)
	*r_key_fpr = strdup (keygen.key_fpr? keygen.key_fpr : "");
    return err;
}


/* Change the pin from @pin to @pin_new.
   Return value: 0 on success. */
gpgme_error_t 
GpgCardEdit::changePIN (int _type)
{
    gpgme_error_t err;

    if (!this->pin_new)
	return gpg_error (GPG_ERR_INV_ARG);

    if (!_type) {
	if (this->pin && this->pin_new)
	    this->type = GPG_EDITCARD_CHUPIN;
	else if (this->pin && this->admin_pin)
	    this->type = GPG_EDITCARD_CHAPIN;
	else
	    this->type = GPG_EDITCARD_UNBPIN;
    }
    else
	this->type = _type;
    /* check if the user provided the needed PIN. */
    if ((this->type == GPG_EDITCARD_CHUPIN && !this->pin) ||
	(this->type == GPG_EDITCARD_CHAPIN && !this->admin_pin))
	return gpg_error (GPG_ERR_INV_ARG);

    err = gpg_card_edit (ctx, this);
    return err;
}


gpgme_error_t
GpgCardEdit::updateName (const char *given, const char *sur)
{
    gpgme_error_t err;

    if (!this->admin_pin)
	return gpg_error (GPG_ERR_INV_OBJ);

    this->type = GPG_EDITCARD_NAME;
    this->edit.surname = sur;
    this->edit.givenname = given;

    err = gpg_card_edit (ctx, this);
    return err;
}
    
gpgme_error_t
GpgCardEdit::updateURL (const char *_url)
{
    gpgme_error_t err;

    if (!this->admin_pin)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITCARD_KEYURL;
    this->edit.keyurl = _url;

    err = gpg_card_edit (ctx, this);
    return err;
}


gpgme_error_t
GpgCardEdit::updateLogin (const char *login)
{
    gpgme_error_t err;

    if (!this->admin_pin)
	return gpg_error (GPG_ERR_INV_OBJ);

    this->type = GPG_EDITCARD_LOGIN;
    this->edit.login = login;

    err = gpg_card_edit (ctx, this);
    return err;
}

gpgme_error_t
GpgCardEdit::updateSex (char sex)
{
    gpgme_error_t err;

    if (!this->admin_pin)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITCARD_SEX;
    this->edit.sex = sex;

    err = gpg_card_edit (ctx, this);
    return err;
}

gpgme_error_t
GpgCardEdit::updateLanguage (const char *lang)
{
    gpgme_error_t err;

    if (!this->admin_pin)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITCARD_LANG;
    this->edit.lang = lang;

    err = gpg_card_edit (ctx, this);
    return err;
}


/* Fetch the key specified by the url stored on the card.
   Return value: 0 on success. */
gpgme_error_t
GpgCardEdit::fetchKey (void)
{
    if (!this->pin)
	return gpg_error (GPG_ERR_INV_OBJ);

    return 0;
}


/* Dispatcher for the various commands. @cmd is the card command ID
   and @arg1 and @arg2 the actual values for the command.
   Return value: 0 on success. */
gpgme_error_t
GpgCardEdit:: doCmd (int cmd, const char *arg1, const char *arg2)
{
    switch (cmd) {
    case GPG_EDITCARD_NAME:
	return updateName (arg1, arg2); /* given, surname */

    case GPG_EDITCARD_LANG:
	return updateLanguage (arg1); /* lang */

    case GPG_EDITCARD_SEX:
	return updateSex (*arg1); /* sex */

    case GPG_EDITCARD_KEYURL:
	return updateURL (arg1); /* url */

    case GPG_EDITCARD_LOGIN:
	return updateLogin (arg1); /* login */

    default:
	return gpg_error (GPG_ERR_INV_VALUE);
    }

    return 0;
}


/* Set the result for the executed operation.
   The values will be ORed. */
void
GpgCardEdit::setResult (int res)
{
    result |= res;
}


/* Return the result of the executed operation. */
int
GpgCardEdit::getResult (void)
{
    return result;
}


/* Reset the object state. */
void
GpgCardEdit::reset (void)
{
    cnt = 0;
}


/* Return card command type */
int
GpgCardEdit::getType (void)
{
    return type;
}
