/* wptHTTP.cpp - Generic HTTP support
 *      Copyright (C) 2004, 2005, 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <ctype.h>
#include <errno.h>

#include "wptHTTP.h"
#include "wptTypes.h"
#include "wptErrors.h"

#define free_if_malloc(buf) if ((buf)) { free ((buf)); (buf) = NULL; }

/* Empty constructur to allow advanced requests. */
wHTTP::wHTTP (void)
{
    reset ();
}


/* Constructur to allow alternative syntax. */
wHTTP::wHTTP (const char *_host, int _port, const char *_url)
{
    reset ();
    this->host = strdup (_host);
    this->port = _port;
    this->url = strdup (_url);
    if (!sendRequest (_host, _port, _url))
	parseResponse (&statcode);

}


/* Standard constructur. */
wHTTP::wHTTP (const char *_url)
{
    reset ();
    extractHostInfo (_url, &this->host, &this->url);
    /* XXX: if the connection fails, return the error code to the user. */
    if (!sendRequest (this->host, this->port, this->url))
	parseResponse (&statcode);
}


/* Reset object contents for first use. */
void
wHTTP::reset (void)
{
    ver = 1;
    port = 80;
    host = url = NULL;
    req_headers = resp_headers = NULL;
    fd = statcode = 0;
    method = HTTP_GET;
    nleft = -1;
    error = 0;
}


/* Perform a HTTP 'HEAD' request. */
int
wHTTP::head (const char *_url)
{
    free_if_malloc (host);
    free_if_malloc (url);

    extractHostInfo (_url, &this->host, &this->url);
    method = HTTP_HEAD;

    if (!sendRequest (host, port, url))
	parseResponse (&statcode);

    return 0;
}


/* Perform a HTTP 'GET' request. */
int
wHTTP::get (const char *_url)
{
    free_if_malloc (host);
    free_if_alloc (url);

    extractHostInfo (_url, &this->host, &this->url);

    method = HTTP_GET;
    if (!sendRequest (this->host, this->port, this->url))
	parseResponse (&statcode);

    return 0;
}


/* Return HTTP status code. */
int 
wHTTP::getStatusCode (void)
{
    return statcode;
}


/* Return MIME content type. */
const char*
wHTTP::getContentType (void)
{
    const char *type = NULL;

    findHeader (resp_headers, "Content-Type", &type);
    return type;
}


/* Return content length. */
unsigned int
wHTTP::getContentLength (void)
{
    const char *len = NULL;

    if (findHeader (resp_headers, "Content-Length", &len))
	return strtoul (len, NULL, 10);
    return 0;
}


void
wHTTP::addHeader (http_head_t *root, const char *val)
{
    http_head_t n, t;
    
    t = (http_head_t)calloc (1, sizeof *t+strlen (val)+2);
    if (!t)
	BUG (0);
    strcpy (t->d, val);

    if (!*root)
	*root = t;
    else {
	for (n = *root; n->next; n=n->next)
	    ;
	n->next = t;
    }
}


bool
wHTTP::findHeader (http_head_t root, 
		   const char *name, const char **val)
{
    http_head_t n;
    char *p;
    
    *val = NULL;
    for (n = root; n; n = n->next) {
	if (strlen (n->d) >= strlen (name) &&
	    strstr (n->d, name)) {
	    p = strchr (n->d, ':');
            *val = p? n->d + (p - n->d + 1 + 1) : NULL;
	    return true;
        }
    }
    return false;
}


int
wHTTP::connect (const char *_host, int _port)
{
    struct hostent *hp;
    struct sockaddr_in srv;
    unsigned long addr = 0;
    int i = 1;

    memset (&srv, 0, sizeof srv);
    srv.sin_port = htons ((unsigned short)_port);
    srv.sin_family = AF_INET;
    
    if (isalpha (*_host)) {
	hp = gethostbyname (_host);
	if (!hp) {
	    this->error = (int)WSAGetLastError ();
	    return WPTERR_WINSOCK_CONNECT;
	}
	memcpy (&srv.sin_addr, hp->h_addr, hp->h_length);
    }
    else {
	addr = inet_addr (_host);
	if (addr == INADDR_NONE) {
	    this->error = (int)WSAGetLastError ();
	    return WPTERR_WINSOCK_CONNECT;
	}
	memcpy (&srv.sin_addr, &addr, sizeof addr);
    }
	    
    fd = socket (AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
	this->error = (int)WSAGetLastError ();
	return WPTERR_WINSOCK_SOCKET;
    }

    if (setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, 
		    (const char*)&i, sizeof i)) {
	closesocket (fd);
	fd = 0;
	this->error = (int)WSAGetLastError ();
	return WPTERR_WINSOCK_SOCKET;
    }
    
    if (::connect (fd, (struct sockaddr *)&srv, sizeof srv)) {
	closesocket (fd);
	fd = 0;
	this->error = (int)WSAGetLastError ();
	return WPTERR_WINSOCK_CONNECT;
    }

    return 0;
}


int
wHTTP::isDataAvailable (int _fd)
{
    struct timeval tv;
    fd_set inp;
    int n;

    FD_ZERO (&inp);
    FD_SET (_fd, &inp);
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    n = select (_fd+1, &inp, NULL, NULL, &tv);
    if (n && FD_ISSET (_fd, &inp))
        return n;
    return 0;
}


int
wHTTP::readLine (char *buf, unsigned int nbuf,
                 int nonblock, int *nn, int *eof)
{
    char c;
    int n, i;
    
    if (nn)
	*nn = 0;
    if (eof)
        *eof = 0;
    i = 0;
    do {
        if (nonblock == 1 && isDataAvailable (fd) == 0) {
            buf[i++] = '\0';
            i = -1;
            break;
        }
	n = recv (fd, &c, 1, 0);
	if (n == -1)
	    break;
	if (n == 0 || nbuf == 0 || c == '\n') {
	    if (n == 0) {
                if (eof)
                    *eof = 1;
                buf[i++] = '\0';
	    }
            else {
                buf[i++] = c;
                buf[i] = '\0';
            }
            break;
	}
        else {
            buf[i++] = c;
            nbuf--;
        }
    } while (n > 0 && i != -1);
    
    if (nn)
	*nn = i;
    return 0;
}



/* Extract the host from the given url @url. Return the host in
   @host and also return the resource part of the url in @new_url. */
int
wHTTP::extractHostInfo (const char *_url, char **_host, char **new_url)
{
    char *p;
    char tmpbuf[512];

    *_host = NULL;
    *new_url = NULL;

    /* XXX: do not use static buffers. */
    memset (tmpbuf, 0, sizeof (tmpbuf));
    strncpy (tmpbuf, _url, 512);
    
    p = "http://";
    if (strlen (_url) < 10 || strncmp (_url, p, 7))
        return WPTERR_GENERAL;
    _url += strlen (p);
    p = strtok (tmpbuf+7, "/");
    if (!p)
        return WPTERR_GENERAL;
    *_host = strdup (p);
    p = strchr (_url, '/');
    if (!p) {
	free_if_malloc (*_host);
        return WPTERR_GENERAL;
    }
    *new_url = strdup (p);
    return 0;
}


/* Add a request header. */
int 
wHTTP::addRequestHeader (const char *name, const char *val)
{
    char *p;
    char *fmt = "%s: %s";

    p = (char*) malloc (strlen (name)+strlen (val)+strlen (fmt)+1);
    sprintf (p, fmt, name, val);
    addHeader (&req_headers, p);
    free_if_malloc (p);

    return 0;
}


void 
wHTTP::setVersion (int _ver)
{
    if (_ver > 9)
	_ver = 0;
    this->ver = _ver;
}


int
wHTTP::addRequestHeader (const char *name, unsigned int val)
{
    char buf[32];

    sprintf (buf, "%lu", (DWORD)val);
    return addRequestHeader (name, buf);
}


/* Prepare the request resource @url and send 
   it to host @host (port @port). */
int
wHTTP::sendRequest (const char *_host, int _port, const char *_url)
{
    const char *id[] = {"GET", "HEAD", "PUT", "POST"};
    http_head_t h;
    const char *h_head;
    char *p;
    int n;
    int rc;

    if (!this->fd) {
	rc = connect (_host, _port);
	if (rc)
	    return rc;
    }

    if (*_url == '/')
    	url++;
    if (_url == NULL)
    	_url = "";

    addRequestHeader ("Host", _host);
    if (ver < 1)
	addRequestHeader ("Connection", "close");

    n = 0;
    for (h = req_headers; h; h = h->next)
	n += strlen (h->d) + 2 + 1;

    h_head = "%s /%s HTTP/1.%d\r\n";
    p = (char*)calloc (1, strlen (id[method]) + strlen (h_head)
			  + strlen (url) + n + 2 + 1 + 4);
    if (!p)
	BUG (0);
    sprintf (p, h_head, id[method], url, ver);
    for (h = req_headers; h; h = h->next) {
	strcat (p, h->d);
	strcat (p, "\r\n");
    }
    strcat (p, "\r\n");
    printf ("p='%s'\n", p);
    n = send (fd, p, strlen (p), 0);
    free_if_malloc (p);
    return n > 0? 0 : WPTERR_GENERAL;
}


/* Parse all response resp_headers. */
int
wHTTP::parseHeaders (http_head_t *r_head)
{
    char buf[300];
    int nn;
    int rc;
    
    if (!r_head)
	return WPTERR_GENERAL;

    do  {
	rc = readLine (buf, 299, 1, &nn, NULL);
	if (rc)
	    return rc;
	if (nn == 2)
	    break; /* reach empty line */
	addHeader (r_head, buf);
    } while (rc == 0);
    return 0;
}


/* Read data from the response and write it to @out. */
int
wHTTP::readData (FILE *out)
{
    const char *val;
    char buf[1024+1];
    int nlen = 0, nn = 0, n, eof=0;
    int rc = 0;
    
    if (this->fd == 0 || this->resp_headers == NULL)
	return -1;

    nlen = getContentLength ();
    if (nlen == 0) {
        if (!findHeader (resp_headers, "Connection", &val) ||
	    strnicmp (val, "close", 4))
            return WPTERR_GENERAL;
    }

    val = getContentType ();
    if (!val)
        return WPTERR_GENERAL;
    if (strnicmp (val, "text", 4)) { /* binary */
	do {
	    n = recv (fd, buf, nlen > 1024? 1024 : nlen, 0);
	    if (n == -1)
		return SOCKET_ERROR;
	    if (n == 0)
		break;
	    nlen -= n;
	    fwrite (buf, 1, n, out);
	} while (nlen > 0);
	return 0;
    }

    do {
	rc = readLine (buf, 1024, 1, &n, &eof);
	if (rc)
	    return rc;
	if (n > 0)
	    fwrite (buf, 1, n, out);
        if (nlen > 0) {
            nn += n;
            if (nlen == nn)
                break;
        }
    } while (eof == 0);

    return 0;
}


/* Parse the HTTP response line. */
int
wHTTP::parseResponse (int *_statcode)
{
    http_head_t n;
    const char *tmp, *p;
    char buf[300];
    int code = 0, nn = 0;
    int rc;

    *_statcode = 0;
    rc = readLine (buf, 299, 1, &nn, NULL);
    if (rc)
	return rc;

    tmp = "HTTP/1.";
    if (strlen (buf) < 8 || strncmp (buf, tmp, strlen (tmp)))
	return WPTERR_GENERAL;
    p = buf;
    p += strlen (tmp)+1 + 1; /* skip HTTP/1.x and WS */
    *_statcode = atoi (p);
    if (tmp == 0)
	return WPTERR_GENERAL;
    p += 3 + 1; /* number + WS */

    if (code == HTTP_STAT_400 ||            
	code == HTTP_STAT_403 ||
	code == HTTP_STAT_404 ||
	code == HTTP_STAT_405)
	return WPTERR_GENERAL;

    while (resp_headers) {
	n = resp_headers->next;
	free (resp_headers);
	resp_headers = n;
    }
    resp_headers = NULL;
    rc = parseHeaders (&resp_headers);
    if (rc)
	return rc;
    return 0;
}


/* Destroy HTTP object. */
wHTTP::~wHTTP (void)
{

    http_head_t h;
    
    while (resp_headers) {
	h = resp_headers->next;
	free (resp_headers);
	resp_headers = h;
    }
    while (req_headers) {
	h = req_headers->next;
	free (req_headers);
	req_headers = h;
    }
    free_if_malloc (url);
    free_if_malloc (host);
    if (fd != 0) {
	closesocket (fd);
	fd = 0;
    }
}


/* Read @buflen bytes from the http stream into the buffer @buf. */
int
wHTTP::read (void *buf, unsigned int buflen)
{
    int n;

    if (this->fd == 0 || this->resp_headers == NULL)
	return -1;
    if (method != HTTP_GET)
	return -1;

    if (nleft == -1)
	nleft = getContentLength ();
    if ((int)buflen > nleft)
	buflen = nleft;
    if (nleft == 0)
	return -1;

    n = recv (fd, (char*)buf, (int)buflen, 0);
    if (n > 0) {
	nleft -= n;
	if (nleft < 0) nleft = 0;
    }
    return n;
}


/* Write the buffer @buf to the http stream. */
int
wHTTP::write (const void *buf, unsigned buflen)
{
    if (this->fd == 0)
	return -1;
    if (method == HTTP_GET || method == HTTP_HEAD)
	return -1;
    return send (fd, (const char*)buf, (int)buflen, 0);
}


/* Return the winsock specific error code. */
int wHTTP::getErrorCode (void)
{
    return error;
}
